/*
 * ar4c-dcdc - comm-avr.c
 *
 * Communication with host
 *
 * Copyright (C) 2014  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <compiler.h>

#ifndef F_CPU
#define F_CPU	1000000UL
#endif
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <comm.h>
#include <crc7kb.h>

#define CSUM_INIT	0xff

/*
 * Protocol:
 *
 *       0                   1                   2                   3
 *       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * MOSI |1|W|C|   addr  |0|    DATA     |0|    DATA     |0|   CHECKSUM  |
 *      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * MISO |     0x00      |      DATA     |      DATA     |    CHECKSUM   |
 *      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */
#define DC_SYNC		0x80
#define DC_WRITE	0x40
#define DC_COMMAND	0x20
#define DC_REGS_MASK	0x1f

#define DC_CMD_POWEROFF0	0x3
#define DC_CMD_POWEROFF1	0xe

#define DC_POWEROFF_MAGIC0	0x5a
#define DC_POWEROFF_MAGIC1	0x3c

#define NUM_REGS	32

uint16_t comm_regs[NUM_REGS];

static u8 comm_command;

static unsigned char spi_byte = 0xff;
static uint16_t spi_data;

void comm_reg_write(unsigned char reg, int value)
{
	cli();
	comm_regs[reg & DC_REGS_MASK] = value;
	sei();
}

void comm_reg_clear(unsigned char reg, int value)
{
	cli();
	comm_regs[reg & DC_REGS_MASK] &= ~value;
	sei();
}

/*
 * We don't need disabling interrupt here we currently not allow
 * modification of registers from interrupt.
 */
int comm_reg_read(unsigned char reg)
{
	return comm_regs[reg];
}

unsigned char comm_get_command(void)
{
	unsigned char ret = comm_command;

	comm_command = 0;
	return ret;
}

static void exec_cmd(u8 *cmd)
{
	u8 c = cmd[0] & 0x0f;
	u8 arg = cmd[1];

	if (c == DC_CMD_POWEROFF0 && arg == DC_POWEROFF_MAGIC0)
		comm_regs[COMM_COMMAND] |= COMM_CMD_POWEROFF0;
	else if (c == DC_CMD_POWEROFF1 && arg == DC_POWEROFF_MAGIC1)
		comm_regs[COMM_COMMAND] |= COMM_CMD_POWEROFF1;
}

ISR(SPI_STC_vect) {
	static u8 rxbuf[4];
	static u8 csum;
	unsigned char rx;
	unsigned char tx = 0x00;

	rx = SPDR;
	if (rx & DC_SYNC) {
		spi_byte = 0;
		rxbuf[0] = rx;
		if ((rx & DC_WRITE) == 0) {
			spi_data = comm_regs[rx & DC_REGS_MASK];
			tx = spi_data >> 8;
		}
	} else {
		switch (spi_byte) {
		case 0:
			tx = spi_data & 0xff;
			break;
		case 1:
			tx = csum;
			break;
		}
		if (spi_byte < 0xff)
			spi_byte++;
		if (spi_byte < sizeof(rxbuf))
			rxbuf[spi_byte] = rx;
	}

	SPDR = tx;

	if (!spi_byte) {
		csum = CSUM_INIT;
		csum = crc7kb_byte(csum, rx);
		csum = crc7kb_byte(csum, tx);
		csum = crc7kb_byte(csum, spi_data & 0xff);
		csum = ~csum << 1;
	}

	if ((spi_byte == 2) && ((rxbuf[0] & 0xf0) == 0xa0))
		exec_cmd(rxbuf);
}

void __init comm_init(void)
{
	SPCR = BIT(SPE) | BIT(SPIE);

	sei();
}
